use std::collections::binary_heap::BinaryHeap;
use std::ops::DerefMut;
use std::sync::{Arc, Mutex};
use std::sync::mpsc::{channel, Receiver, Sender};
use std::thread::JoinHandle;
use std::time::Duration;
use std::vec::Vec;

use ::task::Task;
use ::thread::Thread;
use ::thread_data::ThreadData;
use ::thread_state::ThreadState;



///
pub struct Scheduler
{
   /// The minimum amount of compute threads that the
   /// scheduler will keep alive.
   minimum_thread_amount: u64,

   /// The maximum amount of compute threads that the
   /// scheduler will have alive at one given time.
   maximum_thread_amount: u64,

   /// Specify the maximum amount of time that a thread
   /// can stay idle before it is marked to be killed.
   maximum_thread_idle_time: Duration,

   /// The threads created to handle
   /// non-blocking computation tasks.
   compute_threads: Vec<ThreadData>,

   /// The threads created to handle blocking tasks.
   blocking_threads: Vec<ThreadData>,

   /// The main task queue that threads can use to process tasks.
   task_queue: Arc<Mutex<BinaryHeap<Task>>>
}



impl Scheduler
{
   /// 
   pub fn new(min_threads: Option<u64>, max_threads: Option<u64>) -> Scheduler
   {
      let min_count: u64;
      let max_count: u64;

      // See if there is a minimum number of threads that
      // the scheduler is supposed to keep around.
      match min_threads
      {
         Some(count) =>
         {
            min_count = count;
         }

         None =>
         {
            min_count = 0;
         }
      }

      // Get the maximum amount of threads that the
      // scheduler can create. This is for compute threads
      // since I/O threads will block.
      match max_threads
      {
         Some(count) =>
         {
            // Set the maximum number of threads to be the desired
            // amount.
            max_count = count;
         }

         None =>
         {
            // We don't know how many threads we can create.
            // Do the ideal thing and create 1 thread per core.
            max_count = 1;
         }
      }

      debug!("Creating Scheduler with {} max threads.", max_count);

      Scheduler
      {
         minimum_thread_amount: min_count,
         maximum_thread_amount: max_count,
         maximum_thread_idle_time: Duration::new(5, 0),
         compute_threads: Vec::with_capacity(max_count as usize),
         blocking_threads: Vec::with_capacity(max_count as usize),
         task_queue: Arc::new(Mutex::new(BinaryHeap::new()))
      }
   }

   /// Queues a task to be processed.
   pub fn queue_task(&mut self, task: Task)
   {
      // Just add the task to the queue.
      match self.task_queue.lock()
      {
         Ok(ref mut guard) =>
         {
            guard.deref_mut().push(task);
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }
   }

   /// Processing the currently queued tasks.
   ///
   /// If this is not a single iteration, then it will block
   /// the current thread until all tasks are finished processed.
   /// This does not count immortal tasks.
   pub fn process_tasks(&mut self, is_single_iteration: bool)
   {
      let mut continue_processing: bool;
      let mut new_thread_count: u64;

      // Continue to process the tasks as long as possible.
      continue_processing = true;
      while continue_processing == true
      {
         debug!("There are currently {} tasks to process.",
                self.get_task_count());
         debug!("Currently running {} out of {} threads.",
                self.compute_threads.len(), self.maximum_thread_amount);

         // Determine how many threads need to be created.
         new_thread_count = self.determine_new_thread_amount();

         // Create the threads that we determined that we needed
         // to process the current tasks.
         self.spawn_compute_threads(new_thread_count);

         // Check to see if this was to be a single iteration
         // of the task scheduler.
         if is_single_iteration == true
         {
            // This was only a single iteration,
            // so stop the processing.
            continue_processing = false;
         }
//         else
//         {
            // Let the thread sleep before
            // handling the next iteration.
            ::std::thread::sleep(Duration::new(0, 100));
//         }
      }
   }

   /// Get the amount of tasks in the queue.
   fn get_task_count(&mut self) -> u64
   {
      // Get the size of the task queue.
      match self.task_queue.lock()
      {
         Ok(ref mut guard) =>
         {
            return guard.deref_mut().len() as u64;
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }
   }

   /// This will calculate the new amount of compute threads
   /// that need to be created.
   fn determine_new_thread_amount(&mut self) -> u64
   {
      let task_count: u64;
      let thread_count: u64;
      let potential_threads: u64;
      let mut desired_threads: u64;
      let mut available_threads: u64;
      let mut new_thread_count: u64;

      // Determine the amount of tasks that need to be processed.
      task_count = self.get_task_count();
      debug!("There are currently {} compute tasks to process.", task_count);

      // Determine how many threads we currently have available.
      thread_count = self.compute_threads.len() as u64;
      debug!("Currently running {} out of {} compute threads.",
             thread_count, self.maximum_thread_amount);

      // Determine how many of these threads currently are
      // not processing anything.
      available_threads = 0u64;
      for thread_data in self.compute_threads.iter()
      {
         match thread_data.state
         {
            ThreadState::Idle =>
            {
               available_threads += 1u64;
            }

            ThreadState::Starting =>
            {
               available_threads += 1u64;
            }

            _ =>
            {
            }
         }
      }

      // Calculate how many threads we would need to
      // process all the tasks in our queue if each
      // task was given its own thread.
      println!("task_count: {}\navailable_threads: {}", task_count, available_threads);
      desired_threads = 0u64;
      if task_count > available_threads
      {
         desired_threads = task_count - available_threads;
      }
      new_thread_count = desired_threads;

      // Calculate how many potential threads can be created.
      potential_threads = self.maximum_thread_amount - thread_count;

      // We are limited by a hard ceiling of runnable threads.
      // Make sure we have as many as we can to process
       // the currently queued tasks.
      if potential_threads < desired_threads
      {
         // Make sure we have atleast the maximum number
         // of threads aloud.
         new_thread_count = self.maximum_thread_amount - thread_count;
      }

      // Return the determined amount of threads to create.
      new_thread_count
   }

   ///
   fn spawn_compute_threads(&mut self, amount: u64)
   {
      let mut shutdown_channel: (Sender<bool>, Receiver<bool>);
      let mut state_channel: (Sender<ThreadState>, Receiver<ThreadState>);
      let mut join_handle: JoinHandle<()>;
      let mut data: ThreadData;
      let mut new_thread: Thread;

      // Loop through and create all the required threads.
      debug!("Spawning {} compute threads", amount);
      for _i in 0..amount
      {
         // Create the channel to shutdown the thread.
         shutdown_channel = channel::<bool>();

         // Create the channel to retrieve the
         // status of the thread.
         state_channel = channel::<ThreadState>();

         // Create a new Thread.
         new_thread = Thread::new(self.task_queue.clone(),
                                  shutdown_channel.1, state_channel.0);
         join_handle = ::std::thread::spawn(move ||
            {
               new_thread.process();
            }
         );

         // Create a new set of data for the
         // thread we want to create.
         data = ThreadData::new(join_handle, shutdown_channel.0,
                                state_channel.1);

         // Add the new ThreadData to the set
         // of compute threads.
         self.compute_threads.push(data);
      }
   }
}

impl Drop for Scheduler
{
   fn drop(&mut self)
   {
      debug!("Destroying scheduler.");

      // Stop any threads that are running.
      debug!("Stopping threads.");
      for thread_data in self.compute_threads.iter_mut()
      {
         // Stop the thread.
         match thread_data.shutdown_sender.send(true)
         {
            Ok(_) =>
            {
            }

            Err(error) =>
            {
               warn!("{}", error);
            }
         }
      }

      // Wait about 5 seconds to make sure all
      // the threads have been stopped.
      debug!("Giving threads a chance to end.");
      ::std::thread::sleep(Duration::new(5, 0));
   }
}
